from werkzeug.exceptions import BadRequest

from flask_camp.models import User

from tests.unit_tests.utils import BaseTest


def on_user_update(user):
    user.name = "not the good name"
    raise BadRequest()


class Test_AutomaticRollback(BaseTest):
    rest_api_kwargs = {
        "on_user_update": on_user_update,
    }

    def test_main(self, user):
        def check(r):
            updated_user = self.api.database.session.query(User).get(user.id)
            assert updated_user.name != "not the good name"
            return r

        self.app.after_request(check)

        self.login_user(user)
        self.modify_user(user, expected_status=400)



from flask import request
from flask_login import current_user
from sqlalchemy.exc import IntegrityError
from werkzeug.exceptions import Forbidden, NotFound, BadRequest

from flask_camp._schemas import schema
from flask_camp._utils import current_api
from flask_camp._services._security import allow
from flask_camp.models._user import User as UserModel


rule = "/user/<int:user_id>"


@allow("anonymous", "authenticated", allow_blocked=True)
def get(user_id):
    """Get an user"""
    user = UserModel.get(id=user_id)

    if user is None:
        raise NotFound()

    include_personal_data = False

    if current_user.is_authenticated:
        if user.id == current_user.id:
            include_personal_data = True
        elif current_user.is_admin:
            include_personal_data = True

    return {
        "status": "ok",
        "user": user.as_dict(include_personal_data=include_personal_data),
    }


@allow("authenticated", allow_blocked=True)
@schema("modify_user.json")
def put(user_id):
    """Modify an user"""
    data = request.get_json()

    if user_id != current_user.id:
        if not current_user.is_moderator and not current_user.is_admin:
            raise Forbidden("You can't modify this user")

        if "comment" not in data:
            raise BadRequest("comment is missing")

    current_api.validate_user_schema(data["user"])

    user = UserModel.get(id=user_id)

    if user is None:
        raise NotFound()

    if "name" in data["user"]:
        _update_name(user, data["user"]["name"])

    if "roles" in data["user"]:
        _update_roles(user, data["user"]["roles"])

    if "blocked" in data["user"]:
        _update_blocked(user, data["user"]["blocked"])

    if "ui_preferences" in data["user"]:
        user.ui_preferences = data["user"]["ui_preferences"]

    password = data["user"].get("password", None)
    token = data["user"].get("token", None)
    new_password = data["user"].get("new_password", None)
    email = data["user"].get("email", None)

    if new_password is not None or email is not None:
        if not user.check_auth(password=password, token=token):
            raise Forbidden()

        if new_password is not None:
            user.set_password(new_password)

        if email is not None:
            user.set_email(email)
            user.send_email_change_mail()

    # TODO : on_user_update: old and new version
    current_api.on_user_update(user)

    try:
        current_api.database.session.commit()
    except IntegrityError as e:
        raise BadRequest("Name or email already exists") from e

    return {"status": "ok", "user": user.as_dict(include_personal_data=current_user.id == user.id)}


def _update_blocked(user, blocked):

    if blocked == user.blocked:
        return

    if not current_user.is_moderator:
        raise Forbidden("Only moderator can change names")

    user.blocked = blocked

    current_api.add_log(action="block" if blocked else "unblock", target_user=user)
    current_api.on_user_block(user)


def _update_name(user, name):
    name = UserModel.sanitize_name(name)

    if name == user.name:
        return

    if user.id != current_user.id:
        current_api.add_log(action="rename", target_user=user)

    user.name = name


def _update_roles(user, roles):
    roles = sorted(roles)

    if roles == sorted(user.roles):
        return

    if not current_user.is_admin:
        raise Forbidden("Only admin can change roles")

    for new_role in roles:

        if new_role not in current_api.user_roles:
            raise BadRequest(f"'{new_role}' doesn't exists. Possible roles are {sorted(current_api.user_roles)}.")

        if new_role not in user.roles:
            current_api.add_log(action=f"add_role {new_role}", target_user=user)

    for old_role in user.roles:
        if old_role not in roles:
            current_api.add_log(action=f"remove_role {old_role}", target_user=user)

    user.roles = roles



from flask_camp.models import User
from tests.unit_tests.utils import BaseTest


class Test_UserLogin(BaseTest):
    def test_normalization(self):
        with self.app.app_context():
            user = User.create(name="  CaPiTaL  ", email="  CaPiTaL@example.com  ", password="x")

        assert user.name == "capital"
        assert user._email_to_validate == "capital@example.com"

    def test_name(self, user):
        self.put("/user/login", json={"name_or_email": user.name, "password": "password"}, expected_status=200)

    def test_name_capitalization(self, user):
        self.put("/user/login", json={"name_or_email": user.name.upper(), "password": "password"}, expected_status=200)
        self.put("/user/login", json={"name_or_email": user.name.lower(), "password": "password"}, expected_status=200)

    def test_email(self, user):
        self.put("/user/login", json={"name_or_email": user._email, "password": "password"}, expected_status=200)

    def test_email_capitalization(self, user):
        self.put(
            "/user/login", json={"name_or_email": user._email.upper(), "password": "password"}, expected_status=200
        )
        self.put(
            "/user/login", json={"name_or_email": user._email.lower(), "password": "password"}, expected_status=200
        )

    def test_login_errors(self, user):
        r = self.login_user("not_the_name", expected_status=401)
        assert r.json["description"] == "User does not exists, or password is wrong"

        r = self.login_user(user, password="not the password", expected_status=401)
        assert r.json["description"] == "User does not exists, or password is wrong"

        self.put("/user/login", json={"name_or_email": user.name}, expected_status=400)



import pytest

from flask_camp import RestApi
from flask_camp.exceptions import ConfigurationError
from tests.unit_tests.utils import BaseTest


class Test_UrlPrefix(BaseTest):
    rest_api_kwargs = {"url_prefix": "/v7"}

    def test_main(self):
        class Route:
            rule = "/hello"

            @self.api.allow("anonymous")
            def get(self):
                return "hello"

        self.api.add_modules(self.app, Route(), url_prefix="")

        self.get("/v7/", expected_status=200)
        self.get("/hello", expected_status=200)


def test_not_leading_slash():
    with pytest.raises(ConfigurationError):
        RestApi(url_prefix="v7")

    with pytest.raises(ConfigurationError):
        api = RestApi()
        api.add_modules(None, url_prefix="v7")


def test_tailing_slash():
    with pytest.raises(ConfigurationError):
        RestApi(url_prefix="/v7/")

    with pytest.raises(ConfigurationError):
        api = RestApi()
        api.add_modules(None, None, url_prefix="/v7/")



from contextlib import contextmanager

from tests.unit_tests.utils import BaseTest

NOT_YET_KNOWN = {}


class Hooks:

    calls = []

    @staticmethod
    @contextmanager
    def expect_single_call(document, old_version, new_version):
        Hooks.calls = [
            (document, old_version, new_version),
        ]

        yield

        assert len(Hooks.calls) == 0, "I was expecting more calls"

    @staticmethod
    @contextmanager
    def expect(calls):
        Hooks.calls = calls

        yield

        assert len(Hooks.calls) == 0, "I was expecting more calls"

    @staticmethod
    def on_document_save(document, old_version, new_version):

        if len(Hooks.calls) == 0:
            raise Exception("I should not have been called")

        expected_document, expected_old_version, expected_new_version = Hooks.calls.pop(0)

        assert document is not None

        if new_version is not None:
            assert new_version is document.last_version

        if new_version is not None and old_version is not None:
            assert new_version.id != old_version.id

        if expected_document is not None:
            assert document.id == expected_document["id"]

        if expected_old_version is not None:
            assert old_version is not None
            assert old_version.id == expected_old_version["version_id"]
        else:
            assert old_version is None

        if expected_new_version is NOT_YET_KNOWN:
            assert new_version is not None
        elif expected_new_version is not None:
            assert new_version is not None
            assert new_version.id == expected_new_version["version_id"]
        else:
            assert new_version is None


class Test_UserCreation(BaseTest):
    rest_api_kwargs = {"on_document_save": Hooks.on_document_save}

    def test_regular(self, moderator, admin):
        self.login_user(moderator)

        with Hooks.expect_single_call(document=None, old_version=None, new_version=NOT_YET_KNOWN):
            doc_v1 = self.create_document().json["document"]

        with Hooks.expect_single_call(document=doc_v1, old_version=doc_v1, new_version=NOT_YET_KNOWN):
            doc_v2 = self.modify_document(doc_v1, data=42).json["document"]

        # on_document_save is not called
        self.protect_document(doc_v1)
        self.unprotect_document(doc_v1)

        # hide not the last version => no call
        self.hide_version(doc_v1)
        self.unhide_version(doc_v1)

        with Hooks.expect_single_call(document=doc_v1, old_version=doc_v2, new_version=doc_v1):
            self.hide_version(doc_v2)

        with Hooks.expect_single_call(document=doc_v1, old_version=doc_v1, new_version=doc_v2):
            self.unhide_version(doc_v2)

        with Hooks.expect_single_call(document=doc_v1, old_version=doc_v2, new_version=NOT_YET_KNOWN):
            doc_v3 = self.modify_document(doc_v2, data=43).json["document"]

        self.login_user(admin)

        # not the last version, no call
        self.delete_version(doc_v2)

        with Hooks.expect_single_call(document=doc_v1, old_version=doc_v3, new_version=doc_v1):
            self.delete_version(doc_v3)

    def test_merge(self, moderator):
        self.login_user(moderator)

        with Hooks.expect_single_call(document=None, old_version=None, new_version=NOT_YET_KNOWN):
            doc_1 = self.create_document().json["document"]

        with Hooks.expect_single_call(document=None, old_version=None, new_version=NOT_YET_KNOWN):
            doc_2 = self.create_document().json["document"]

        with Hooks.expect_single_call(document=None, old_version=None, new_version=NOT_YET_KNOWN):
            doc_3 = self.create_document().json["document"]

        with Hooks.expect_single_call(document=doc_1, old_version=doc_1, new_version=None):
            self.merge_documents(doc_1, doc_2)

        with Hooks.expect(
            [
                (doc_2, doc_2, doc_3),  # destination, doc_3 is more recent
                (doc_3, doc_3, None),  # merged document
            ]
        ):
            self.merge_documents(doc_3, doc_2)



import re

import pytest

from flask_camp import RestApi, SchemaValidator
from flask_camp.exceptions import ConfigurationError
from tests.unit_tests.utils import BaseTest


def test_error():
    # document schema does not exists
    with pytest.raises(FileNotFoundError) as e:
        RestApi(schemas_directory="tests/unit_tests/schemas/", document_schemas=["notfound.json"])
    assert e.match("File tests/unit_tests/schemas/notfound.json does not exists")

    # user schema does not exists
    with pytest.raises(FileNotFoundError) as e:
        RestApi(schemas_directory="tests/unit_tests/schemas/", user_schema="notfound.json")
    assert e.match("File tests/unit_tests/schemas/notfound.json does not exists")

    # schemas_directory does not exists
    with pytest.raises(FileNotFoundError) as e:
        RestApi(schemas_directory="tests/not_a_dir/")
    assert e.match("tests/not_a_dir/ is not a directory")

    # do not provide schemas_directory
    with pytest.raises(ConfigurationError) as e:
        RestApi(document_schemas=["outing.json"])
    assert e.match("You provide document_schemas wihtout schemas_directory")

    # do not provide schemas_directory
    with pytest.raises(ConfigurationError) as e:
        RestApi(user_schema=["outing.json"])
    assert e.match("You provide user_schema wihtout schemas_directory")

    with pytest.raises(ValueError) as e:
        SchemaValidator("tests/unit_tests/schemas_with_error")
    assert e.match("JSON syntax error in tests/unit_tests/schemas_with_error/outing.json")


class Test_Schemas(BaseTest):
    rest_api_kwargs = {
        "schemas_directory": "tests/unit_tests/schemas",
        "document_schemas": ["outing.json"],
        "user_schema": "user.json",
    }

    def test_basic_document(self, user):

        invalid_data = (
            {"namespace": "outing"},
            {"namespace": "outing", "value": None},
            {"namespace": "outing", "value": 12},
            {"namespace": "outing", "value": "str"},
            {"namespace": "outing", "value": "str", "rating": None},
            {"namespace": "outing", "value": "str", "rating": 12},
            {"namespace": "outing", "value": "str", "rating": "a6"},
        )

        self.login_user(user)

        for data in invalid_data:
            self.create_document(data=data, expected_status=400)

        doc = self.create_document(
            data={"namespace": "outing", "value": "str", "rating": "6a"},
            expected_status=200,
        ).json["document"]

        for data in invalid_data:
            self.modify_document(doc, data=data, expected_status=400)

        self.modify_document(doc, data={"namespace": "outing", "value": "str", "rating": "6b"}, expected_status=200)

        route = self.create_document(data={"namespace": "route"}, expected_status=200).json["document"]
        self.modify_document(route, data=12, expected_status=200)

    def test_basic_user(self):
        invalid_data = (None, 12, {}, {"lang": 12}, {"lang": "sp"})

        for item in invalid_data:
            self.create_user(ui_preferences=item, expected_status=400)

        with self.api.mail.record_messages() as outbox:
            user = self.create_user(ui_preferences={"lang": "fr"}, expected_status=200).json["user"]
            self.validate_email(user, token=re.sub(r"^(.*email_token=)", "", outbox[0].body))

        self.login_user(user)

        for item in invalid_data:
            self.modify_user(user, ui_preferences=item, expected_status=400)

        self.modify_user(user=user, ui_preferences={"lang": "de"}, expected_status=200)



from flask_camp.models import Document, DocumentVersion, User
from tests.unit_tests.utils import BaseTest


class Test_DirectDelete(BaseTest):
    def test_main(self, user):
        with self.app.app_context():
            doc = Document.create("", {}, User.get(user.id))
            assert DocumentVersion.query.count() == 1

            self.api.database.session.commit()

            doc = Document.get(doc.id)
            self.api.database.session.delete(doc)
            self.api.database.session.commit()

            assert DocumentVersion.query.count() == 0

            doc = Document.create("", {}, User.get(user.id))
            assert DocumentVersion.query.count() == 1

            Document.query.filter(Document.id.in_([doc.id])).update({"last_version_id": None})
            self.api.database.session.query(DocumentVersion).filter(DocumentVersion.document_id.in_([doc.id])).delete(
                synchronize_session=False
            )
            self.api.database.session.query(Document).filter(Document.id.in_([doc.id])).delete(
                synchronize_session=False
            )

            self.api.database.session.commit()

            assert Document.query.count() == 0
            assert DocumentVersion.query.count() == 0



import logging
import sys

from flask import Flask
import pytest

from flask_camp import RestApi
from flask_camp.models import User
from flask_camp.views import home


tested_app = Flask(__name__, static_folder=None)
tested_app.config.update({"TESTING": True, "SECRET_KEY": "not very secret", "SQLALCHEMY_TRACK_MODIFICATIONS": False})
tested_api = RestApi(app=tested_app)


logging.basicConfig(format="%(asctime)s [%(levelname)8s] %(message)s")


def pytest_configure(config):
    if config.getoption("-v") > 1:
        logging.getLogger("sqlalchemy").addHandler(logging.StreamHandler(sys.stdout))
        logging.getLogger("sqlalchemy").setLevel(logging.INFO)

    if not config.option.collectonly:

        with tested_app.app_context():
            # Generate some docs. Should be a pre-commit hook ?
            data = home.get()

            order = {
                "GET": 0,
                "POST": 1,
                "PUT": 2,
                "DELETE": 3,
            }

            with open("docs/endpoints.md", mode="w", encoding="utf-8") as f:
                for endpoint in sorted(data):
                    methods = data[endpoint]
                    for method in sorted(methods, key=order.get):
                        infos = methods[method]
                        f.write(f"* `{method.upper()} {endpoint}`: {infos['description']}\n")

            # clean previous uncleaned state
            # do not perform this on collect, editors that automatically collect tests on file change
            # may break current test session

            # why not using tested_api.database.drop_all()?
            # because in some case, a table is not known by the ORM
            # for instance, run test A that define a custom table, stop it during execution (the table is not removed)
            # then run only test B. Table defined in test A is not known

            sql = "SELECT table_name FROM information_schema.tables WHERE table_schema = 'public';"
            rows = tested_api.database.session.execute(sql)
            names = [name for name, in rows]
            if len(names) != 0:
                tested_api.database.session.execute(f"DROP TABLE {','.join(names)} CASCADE;")
                tested_api.database.session.commit()

            tested_api.database.create_all()

        tested_api.memory_cache.flushall()


def _db_add_user(name="name", email=None, password="password", validate_email=True, roles=None):

    instance = User.create(
        name=name,
        password=password,
        email=email if email else f"{name}@site.org",
        roles=roles if isinstance(roles, (list, tuple)) else roles.split(",") if isinstance(roles, str) else [],
    )

    if validate_email:
        instance.validate_email(instance._email_token)

    tested_api.database.session.add(instance)
    tested_api.database.session.commit()

    result = User(
        id=instance.id,
        name=instance.name,
        _email=instance._email,
        _email_to_validate=instance._email_to_validate,
        _email_token=instance._email_token,
        roles=instance.roles,
    )

    return result


@pytest.fixture()
def admin():
    with tested_app.app_context():
        yield _db_add_user(name="admin", roles="admin")


@pytest.fixture()
def moderator():
    with tested_app.app_context():
        yield _db_add_user(name="moderator", roles="moderator")


@pytest.fixture()
def user():
    with tested_app.app_context():
        yield _db_add_user()


@pytest.fixture()
def unvalidated_user():
    with tested_app.app_context():
        yield _db_add_user(validate_email=False)


@pytest.fixture()
def user_2():
    with tested_app.app_context():
        yield _db_add_user("user_2")



import re
from unittest.mock import MagicMock

from tests.unit_tests.utils import BaseTest

hooks = MagicMock()


class Test_OnUserUpdate(BaseTest):
    rest_api_kwargs = {
        "on_user_creation": hooks.on_user_creation,
        "on_user_validation": hooks.on_user_validation,
        "on_user_update": hooks.on_user_update,
        "on_user_block": hooks.on_user_block,
    }

    def test_main(self):

        with self.api.mail.record_messages() as outbox:
            user = self.create_user().json["user"]
            token = re.sub(r"^(.*email_token=)", "", outbox[0].body)

        assert hooks.on_user_creation.called
        assert not hooks.on_user_validation.called
        assert not hooks.on_user_update.called

        hooks.reset_mock()
        self.validate_email(user, token)
        assert not hooks.on_user_creation.called
        assert hooks.on_user_validation.called
        assert not hooks.on_user_update.called

        self.login_user(user)

        hooks.reset_mock()
        self.modify_user(user, password="password", new_password="password")
        assert not hooks.on_user_creation.called
        assert not hooks.on_user_validation.called
        assert hooks.on_user_update.called

        hooks.reset_mock()
        self.modify_user(user, ui_preferences="12")
        assert not hooks.on_user_creation.called
        assert not hooks.on_user_validation.called
        assert hooks.on_user_update.called

        hooks.reset_mock()
        with self.api.mail.record_messages() as outbox:
            self.modify_user(user, password="password", email="new@mail.fr")
            token = re.sub(r"^(.*email_token=)", "", outbox[0].body)

        assert not hooks.on_user_creation.called
        assert not hooks.on_user_validation.called
        assert hooks.on_user_update.called

        hooks.reset_mock()
        self.validate_email(user, token)
        assert not hooks.on_user_creation.called
        assert not hooks.on_user_validation.called
        assert hooks.on_user_update.called

    def test_user_block(self, moderator, user):
        self.login_user(moderator)

        self.block_user(user)
        assert hooks.on_user_block.called

        hooks.reset_mock()
        self.unblock_user(user)
        assert hooks.on_user_block.called
